////////////////////////////////////////////////////////////////////////////////
//
// Filename: 	tonegen.v
//
// Project:	Verilog Tutorial Example file
//
// Purpose:	
//
// Creator:	Dan Gisselquist, Ph.D.
//		Gisselquist Technology, LLC
//
////////////////////////////////////////////////////////////////////////////////
//
// Written and distributed by Gisselquist Technology, LLC
//
// This program is hereby granted to the public domain.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY or
// FITNESS FOR A PARTICULAR PURPOSE.
//
////////////////////////////////////////////////////////////////////////////////
//
//
`default_nettype	none
//
module	tonegen(i_clk, i_reset,
		i_wb_cyc, i_wb_stb, i_wb_we, i_wb_addr, i_wb_data, i_wb_sel,
			o_wb_stall, o_wb_ack, o_wb_data,
		o_audio_ce, o_sample, o_audio_pin);
	//
	// Output types:
	//	0: Output MSB of the sinewave
	//	1: Pulse width modulation
	//	2: Pulse density modulation
	//	default: Delta-Sigma modulation
	//
	parameter	OPT_OUTTYPE = 0;
	parameter	CLOCK_RATE_HZ = 100_000_000;	// 100MHz
	parameter	SAMPLE_RATE_HZ = 48_000;	// 48kHz
	//
	input	wire		i_clk, i_reset;
	//
	input	wire		i_wb_cyc, i_wb_stb, i_wb_we;
	input	wire	[1:0]	i_wb_addr;
	input	wire	[31:0]	i_wb_data;
	input	wire	[3:0]	i_wb_sel;
	output	reg		o_wb_stall;
	output	reg		o_wb_ack;
	output	reg	[31:0]	o_wb_data;
	//
	output	reg		o_audio_ce;
	output	reg	[15:0]	o_sample;
	output	reg		o_audio_pin;
	//

	reg		[31:0]	srate_counter;
	reg		[31:0]	r_frequency_step;
	wire	signed	[15:0]	sin;
	reg		[31:0]	phase;
	reg	signed	[16:0]	r_scale;
	reg	signed	[32:0]	scaled;

	////////////////////////////////////////////////////////////////////////
	//
	// Generate our clock enable
	//
	// Verilator lint_off REALCVT
	localparam [31:0]	CLOCK_STEP = SAMPLE_RATE_HZ * 4.0 * (1<<30)
					* 1.0 / CLOCK_RATE_HZ;
	// Verilator lint_on  REALCVT

	always @(posedge i_clk)
		{ o_audio_ce, srate_counter } <= srate_counter + CLOCK_STEP;


	////////////////////////////////////////////////////////////////////////
	//
	// Calculate the phase of our sine wave
	//
	always @(posedge i_clk)
	if (o_audio_ce)
		phase <= phase + r_frequency_step;

	////////////////////////////////////////////////////////////////////////
	//
	// Look up the value of our sine wave in a table
	//
	sintable
	sintablei(i_clk, i_reset, o_audio_ce, phase[31:20], sin);


	////////////////////////////////////////////////////////////////////////
	//
	// Adjust the sinewave for our desired magnitude
	//
	always @(posedge i_clk)
	if (o_audio_ce)
	begin
		scaled <= sin * r_scale;
		o_sample <= scaled[31:16];
	end

	////////////////////////////////////////////////////////////////////////
	//
	// Generate a usable output
	//
	generate if (OPT_OUTTYPE == 0)
	begin

		always @(*)
			o_audio_pin = o_sample[15];

	end else if (OPT_OUTTYPE <= 2)
	begin // PWM or PDM

		reg [15:0]	pwm_counter;

		always @(posedge i_clk)
			pwm_counter <= pwm_counter + 1;

		if (OPT_OUTTYPE == 1)
		begin // PWM

			always @(posedge i_clk)
				o_audio_pin <= (pwm_counter >= { !o_sample[15],
						o_sample[14:0] });

		end else begin // PDM

			wire	[15:0]	brev_counter;

			genvar k;
			for(k=0; k<16; k=k+1)
			begin

				assign brev_counter[k] = pwm_counter[15-k];

			end

			always @(posedge i_clk)
				o_audio_pin <= (brev_counter >= { !o_sample[15],
						o_sample[14:0] });

		end
	end else begin // Delta Sigma

		reg	[15:0]	ds_counter;

		always @(posedge i_clk)
			{ o_audio_pin, ds_counter }
				<= { !o_sample[15], o_sample[14:0] }+ds_counter;

	end endgenerate

	////////////////////////////////////////////////////////////////////////
	//
	// Handle the bus interface
	//
	always @(*)
		o_wb_stall = 0;

	initial	o_wb_ack = 1'b0;
	always @(posedge i_clk)
		o_wb_ack <= !i_reset && i_wb_stb;

	always @(posedge i_clk)
	if (i_wb_stb && i_wb_we)
	case(i_wb_addr)
	2'b00: r_frequency_step <= i_wb_data;
	2'b01: r_scale <= { 1'b0, i_wb_data[15:0] };
	2'b10: begin end // Add functionality here as you see fit
	2'b11: begin end
	endcase

	always @(posedge i_clk)
	case(i_wb_addr)
	2'b00: o_wb_data <= r_frequency_step;
	2'b01: o_wb_data <= { 15'h0, r_scale };
	2'b10: o_wb_data <= 0; // Add functionality here as you see fit
	2'b11: o_wb_data <= 0;
	endcase

	// Make verilator happy
	// Verilator lint_off UNUSED
	wire	unused;
	assign	unused = &{ 1'b0, i_wb_cyc, i_wb_sel, scaled[32], scaled[15:0] };
	// Verilator lint_on  UNUSED

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
// Formal properties section
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
`ifdef	FORMAL
	localparam	F_LGDEPTH=2;

	wire [F_LGDEPTH-1:0] fwb_nreqs, fwb_nacks, fwb_outstanding;

	localparam	F_LGDEPTH=2;

	fwb_slave #(.AW(2),
		.F_LGDEPTH(F_LGDEPTH),
		.F_MAX_STALL(1),
		.F_MAX_ACK_DELAY(2))
	fwb (i_clk, i_reset,
		i_wb_cyc, i_wb_stb, i_wb_we, i_wb_addr,
	    	i_wb_data, i_wb_sel,
		o_wb_ack, o_wb_stall, o_wb_data, 1'b0,
		fwb_nreqs, fwb_nacks, fwb_outstanding);

	always @(*)
	if (i_wb_cyc)
		assert(fwb_outstanding == (o_wb_ack ? 1:0));

`endif
endmodule
